#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <clib/alib_protos.h>
#include <mui/NList_mcc.h>
#include <mui/NListtree_mcc.h>
#include <proto/intuition.h>
#include <proto/muimaster.h>
#include <proto/utility.h>

#include "extra.h"
#include "SDI_hook.h"

#include "bourriquet.h"
#include "classes.h"
#include "lists.h"
#include "config.h"
#include "debug.h"
#include "locale.h"
#include "main.h"
#include "research.h"
#include "sharing.h"
#include "transfert.h"
#include "prefs_gui.h"
#include "utilities.h"

/***************************************************************************
 Classes prives MUI
***************************************************************************/

struct DumData { long dummy; };
struct BC_Data { struct BodyChunkData *BCD; };
struct MUI_CustomClass *CL_MainWin      = NULL;
struct MUI_CustomClass *CL_BodyChunk    = NULL;

// BC_Dispatcher (BodyChunk)
/*** BC_Dispatcher (BodyChunk) - Subclass of BodyChunk, can load images from files ***/
DISPATCHERPROTO(BC_Dispatcher)
{
   switch(msg->MethodID)
      {
         struct BC_Data *data;
         ULONG useold;

         case OM_NEW:
            {
               struct TagItem *tag, *tags = ((struct opSet *)msg)->ops_AttrList;

               obj = DoSuperNew(cl, obj,
                                  MUIA_FixWidth,    16,
                                  MUIA_FixHeight,   16,
                                  MUIA_InnerBottom, 0,
                                  MUIA_InnerLeft,   0,
                                  MUIA_InnerRight,  0,
                                  MUIA_InnerTop,    0,
                                  TAG_MORE,         tags);

               if(obj)
                  {
                     char fname[SIZE_PATHFILE];

                     useold = FALSE;
                     *fname = 0;
                     while((tag = NextTagItem(&tags)))
                        {
                           switch(tag->ti_Tag)
                              {
                                 case MUIA_Bodychunk_UseOld:
                                    useold = (BOOL)tag->ti_Data;
                                 break;
                                 case MUIA_Bodychunk_File:
                                    if(tag->ti_Data) stccpy(fname, (char *)tag->ti_Data, SIZE_PATHFILE);
                                 break;
                              }
                        }

                     if(*fname)
                        {
                           struct BodyChunkData *BCD;

                           data = INST_DATA(cl, obj);
                           if(useold) BCD = GetBCImage(fname);
                           else BCD = LoadBCImage(fname);

                           if((data->BCD = BCD))
                              {
                                 SetAttrs(obj,
                                            MUIA_FixWidth,              BCD->Width,
                                            MUIA_FixHeight,             BCD->Height,
                                            MUIA_Bitmap_Width,          BCD->Width,
                                            MUIA_Bitmap_Height,         BCD->Height,
                                            MUIA_Bitmap_SourceColors,   BCD->Colors,
                                            MUIA_Bodychunk_Depth,       BCD->Depth,
                                            MUIA_Bodychunk_Body,        BCD->Body,
                                            MUIA_Bodychunk_Compression, BCD->Compression,
                                            MUIA_Bodychunk_Masking,     BCD->Masking,
                                            MUIA_UserData,              useold,
                                            TAG_DONE);
                              }
                        }
                  }
               return(ULONG)obj;
            }
            break;

         case OM_DISPOSE:
            {
               data = INST_DATA(cl, obj);
               get(obj, MUIA_UserData, &useold);
               if(!useold && data->BCD) FreeBCImage(data->BCD);
            }
            break;
      }
   return DoSuperMethodA(cl, obj, msg);
}

///
/// MW_Dispatcher (Main Window)
//  Sous-classe de Windows, utilis pour enlever les sous-fentres en sortie
struct MUIP_MainWindow_CloseWindow { ULONG MethodID; APTR Window; };
DISPATCHERPROTO(MW_Dispatcher)
//static ASM SAVEDS ULONG MW_Dispatcher(REG(a0,struct IClass *cl), REG(a2,Object *obj), REG(a1,Msg msg))
{
   if(msg->MethodID == MUIM_MainWindow_CloseWindow)
      {
         APTR app, win = ((struct MUIP_MainWindow_CloseWindow *)msg)->Window;

         set(win, MUIA_Window_Open, FALSE);
         get(win, MUIA_ApplicationObject, &app);
         DoMethod(app, OM_REMMEMBER, win);
         MUI_DisposeObject(win);
      }
   else return(DoSuperMethodA(cl, obj, (Msg)msg));

   return(0);
}

/// CreateMCC()
// Fonction wrapper pour creer une classe MUI personnalis
static struct MUI_CustomClass *CreateMCC(STRPTR supername, struct MUI_CustomClass *supermcc, int instDataSize, APTR dispatcher)
{
   struct MUI_CustomClass *cl;

   #if defined(__amigaos4__) && defined(WITH_ASMSTUB)
   extern ULONG muiDispatcherEntry(void);

   if((cl = MUI_CreateCustomClass(NULL, supername, supermcc, instDataSize, &muiDispatcherEntry))) cl->mcc_Class->cl_UserData = (ULONG)dispatcher;

   #else
   cl = MUI_CreateCustomClass(NULL, supername, supermcc, instDataSize, dispatcher);
   #endif

   return cl;
}

/*******************************************************************/
/// ExitClasses
//  Enleve les classes MUI personnalises
void ExitClasses(void)
{
   if(CL_MainWin) { MUI_DeleteCustomClass(CL_MainWin); CL_MainWin = NULL; }
   if(CL_BodyChunk) { MUI_DeleteCustomClass(CL_BodyChunk); CL_BodyChunk = NULL; }
}
///
/// InitClasses
//  Initialise les classes MUI personnalises
BOOL InitClasses(void)
{
   if((CL_MainWin      = CreateMCC(MUIC_Window, NULL, sizeof(struct DumData), ENTRY(MW_Dispatcher))))
   if((CL_BodyChunk    = CreateMCC(MUIC_Bodychunk, NULL, sizeof(struct BC_Data), ENTRY(BC_Dispatcher))))
      { return TRUE; }
   return FALSE;
}
///
