#include <stdio.h>
#include <string.h>

#include <clib/alib_protos.h>
#include <libraries/mui.h>
#include <proto/muimaster.h>
#include <proto/exec.h>
#include <proto/utility.h>
#include <proto/intuition.h>
#include <proto/dos.h>

#include "debug.h"

#define SDI_TRAP_LIB 0xFF00 /* SDI prefix to reduce conflicts */

struct SDI_EmulLibEntry
{
  UWORD Trap;
  UWORD pad;
  APTR  Func;
};

#define MAKE_ID(a,b,c,d)  ((ULONG) (a)<<24 | (ULONG) (b)<<16 | (ULONG) (c)<<8 | (ULONG) (d))

struct IntuitionBase	*IntuitionBase	= NULL;
struct Library			*MUIMasterBase	= NULL;
struct Library  		*UtilityBase = NULL;

static Object *app = NULL;
static Object *window = NULL;

static const char *CY_RamDebug_Modes[] =
{
   "RamDebug",
   "File",
   NULL
};

static const char *CY_Dump_Modes[] =
{
   "Executable",
   "File",
   NULL
};

enum { LOG_RAMDEBUG, LOG_FILE };
enum { DUMP_AUTO, DUMP_FILE };

struct LI_Data
{
	char m_register[128];
	char m_function_name[128];
	char m_offset[128];
	char m_line[128];
};

/* Hooks */

APTR ConstructLI_TextFunc(void)
{
  struct LI_Data * data = (struct LI_Data *) REG_A1;
  struct LI_Data *new_entry = (struct LI_Data *) malloc(sizeof(struct LI_Data));
  if (new_entry)
  {
	memcpy(new_entry, data, sizeof(struct LI_Data));
	return (new_entry);
  }
  return (NULL);
}

const struct SDI_EmulLibEntry Gate_ConstructLI_TextFunc =
{
	SDI_TRAP_LIB,
	0,
	(void *) ConstructLI_TextFunc
};

const struct Hook ConstructHook =
{
	{NULL, NULL},
	(HOOKFUNC)&Gate_ConstructLI_TextFunc,
	NULL, NULL
};

void DestructLI_TextFunc(void)
{
  struct LI_Data *entry = (struct LI_Data *)REG_A1;

  if (entry)
	free((void *) entry);
}

const struct SDI_EmulLibEntry Gate_DestructLI_TextFunc =
{
	SDI_TRAP_LIB,
	0,
	(void *) DestructLI_TextFunc
};

const struct Hook DestructHook =
{
	{NULL, NULL},
	(HOOKFUNC)&Gate_DestructLI_TextFunc,
	NULL, NULL
};

void DisplayLI_TextFunc(void)
{
	struct LI_Data *entry = (struct LI_Data *) REG_A1;
	char ** array = (char **) REG_A2;

	if (entry)
	{
		*array++ = (char *) entry->m_register;
		*array++ = (char *) entry->m_function_name;
		*array++ = (char *) entry->m_offset;
		*array   = (char *) entry->m_line;
	}
	else
	{
		*array++ = "Register";
		*array++ = "Function Name";
		*array++ = "Dump Offset";
		*array   = "Dump Line";
	}

}

const struct SDI_EmulLibEntry Gate_DisplayLI_TextFunc =
{
	SDI_TRAP_LIB,
	0,
	(void *) DisplayLI_TextFunc
};

const struct Hook DisplayHook =
{
	{NULL, NULL},
	(HOOKFUNC)&Gate_DisplayLI_TextFunc,
	NULL, NULL
};



struct MUI_CustomClass *cl_debugo;


/*************************************************************************************/

#ifndef	__DECLGATE_H__
#include	"declgate.h"
#endif

#define	DISPATCHERPROTO(name)	static ULONG name##PPC(void); struct EmulLibEntry name = { TRAP_LIB, 0, (void(*)(void))&name##PPC }; static ULONG name ## PPC(void)
#define	DISPATCHERARG				DECLARG_3(a0, struct IClass *, cl, a2, Object *, obj, a1, Msg, msg)


struct DebugData
{
	Object *m_st_executable;
	Object *m_cy_ramdebug_mode;
	Object *m_st_ramdebug_file;
	Object *m_po_ramdebug_file;
	Object *m_cy_dump_mode;
	Object *m_st_dump_file;
	Object *m_po_dump_file;
	Object *m_bt_analyze;
	Object *m_sl_current_hit;
	Object *m_lv_hits;
	Object *m_lst_hits;

	char m_executable[128];

	int  m_ramdebug_mode;
	char m_ramdebugfile[128];

	int  m_dump_mode;
	char m_dumpfile[128];

	int  m_current_hit;

	vector<struct Hit> m_hits;
	vector<struct Func> m_funcs;
};

#define DEBUGBASE			 (TAG_USER + 0x61843728)

#define MUIA_Debug_Executable  (DEBUGBASE + 1)

#define MUIA_Debug_LogMode     (DEBUGBASE + 2)
#define MUIA_Debug_LogFile	   (DEBUGBASE + 3)

#define MUIA_Debug_DumpMode    (DEBUGBASE + 4)
#define MUIA_Debug_DumpFile    (DEBUGBASE + 5)

#define MUIA_Debug_CurrentHit  (DEBUGBASE + 6)


#define MUIV_Debug_Manual      (DEBUGBASE + 10)
#define MUIV_Debug_Auto        (DEBUGBASE + 11)


#define MUIM_Debug_ShowHit	       (DEBUGBASE + 20)
#define MUIM_Debug_PerformAction   (DEBUGBASE + 21)
#define MUIM_Debug_Analyze		   (DEBUGBASE + 22)


static ULONG DebugNew(struct IClass *cl, Object *obj, struct opSet *msg)
{
	Object *st_executable, *cy_ramdebug_mode, *st_ramdebug_file,
		   *cy_dump_mode, *st_dump_file,
		   *bt_analyze,
		   *sl_current_hit,
		   *lv_hits,
		   *lst_hits,
		   *po_executable,
		   *po_ramdebug_file,
		   *po_dump_file;

	obj = (Object * ) DoSuperNew(cl, obj,
					MUIA_Frame, MUIV_Frame_Group,
					Child,
						HGroup, GroupFrameT("Executable"),
							Child, TextObject,
								MUIA_Text_Contents, "Executable path :",
								MUIA_Weight, 0,
							End,
							//Child, st_executable = String("", 127),
							Child,
								po_executable = PopaslObject,
														MUIA_Popstring_Button,
														ImageObject, ImageButtonFrame,
														  MUIA_InputMode, MUIV_InputMode_RelVerify,
														  MUIA_Image_Spec, MUII_PopDrawer,
														  MUIA_Background, MUII_ButtonBack,
														  MUIA_ControlChar, 'n',
														End,
														MUIA_Popstring_String, st_executable = String("",127), End,
						End,

					Child,
						HGroup, GroupFrameT("Log"),
							Child, TextObject,
								MUIA_Text_Contents, "Get from :",
								MUIA_Weight, 0,
							End,
							Child, cy_ramdebug_mode =  CycleObject, MUIA_Cycle_Entries, CY_RamDebug_Modes, MUIA_Cycle_Active, 0, MUIA_Weight, 0, End,
							//Child, st_ramdebug_file = String("", 127),
							Child,
								po_ramdebug_file = PopaslObject,
														MUIA_Popstring_Button,
														ImageObject, ImageButtonFrame,
														  MUIA_InputMode, MUIV_InputMode_RelVerify,
														  MUIA_Image_Spec, MUII_PopDrawer,
														  MUIA_Background, MUII_ButtonBack,
														  MUIA_ControlChar, 'n',
														End,
														MUIA_Popstring_String, st_ramdebug_file = String("",127), End,

						End,

					Child,
						HGroup, GroupFrameT("ASM Dump"),
							Child, TextObject,
								MUIA_Text_Contents, "Get from :",
								MUIA_Weight, 0,
							End,
							Child, cy_dump_mode =  CycleObject, MUIA_Cycle_Entries, CY_Dump_Modes, MUIA_Cycle_Active, 0, MUIA_Weight, 0, End,
							//Child, st_dump_file = String("", 127),
							Child,
								po_dump_file = PopaslObject,
														MUIA_Popstring_Button,
														ImageObject, ImageButtonFrame,
														  MUIA_InputMode, MUIV_InputMode_RelVerify,
														  MUIA_Image_Spec, MUII_PopDrawer,
														  MUIA_Background, MUII_ButtonBack,
														  MUIA_ControlChar, 'n',
														End,
														MUIA_Popstring_String, st_dump_file = String("",127), End,


						End,

					Child,
						bt_analyze =  SimpleButton("Analyze log"),

					Child,
							VGroup, GroupFrameT("Results"),
								Child,
									HGroup,
										Child, TextObject,
											MUIA_Text_Contents, "Display hit :",
											MUIA_Weight, 0,
										End,
										Child, sl_current_hit = SliderObject,
													MUIA_Group_Horiz , TRUE,
													MUIA_Slider_Min  , 1,
													MUIA_Slider_Max  , 10,
													MUIA_Slider_Level, 1,
										End,
									End,

								Child, lv_hits = ListviewObject,
									MUIA_Listview_List, lst_hits = ListObject,
										MUIA_Frame, MUIV_Frame_InputList,
										MUIA_List_Format, "BAR ,W=200 BAR, BAR, BAR",
										MUIA_List_Title, TRUE,
										MUIA_List_DisplayHook, &DisplayHook,
										MUIA_List_ConstructHook, &ConstructHook,
										MUIA_List_DestructHook, &DestructHook,
									End,
								End,
					End,
					TAG_MORE, (struct opSet *) msg->ops_AttrList);

	if (obj)
	{
		struct DebugData *data = (struct DebugData *)INST_DATA(cl, obj);
		struct TagItem *tags,*tag;

		data->m_lv_hits = lv_hits;
		data->m_lst_hits = lst_hits;
		data->m_st_executable = st_executable;
		data->m_cy_ramdebug_mode = cy_ramdebug_mode;
		data->m_st_ramdebug_file = st_ramdebug_file;
		data->m_po_ramdebug_file = po_ramdebug_file;
		data->m_cy_dump_mode = cy_dump_mode;
		data->m_st_dump_file = st_dump_file;
		data->m_po_dump_file = po_dump_file;

		data->m_bt_analyze = bt_analyze;
		data->m_sl_current_hit = sl_current_hit;

		strcpy(data->m_executable, "");
		strcpy(data->m_ramdebugfile, "");
		strcpy(data->m_dumpfile, "");

		data->m_ramdebug_mode = 0;
		data->m_dump_mode = 0;


		/* parse initial taglist */
		for (tags=((struct opSet *)msg)->ops_AttrList;tag=NextTagItem(&tags);)
		{
			switch (tag->ti_Tag)
			{
				case MUIA_Debug_CurrentHit:
					SetAttrs(obj, MUIA_Debug_CurrentHit, tag->ti_Data, TAG_END);
					break;

				case MUIA_Debug_Executable:
					SetAttrs(obj, MUIA_Debug_Executable, tag->ti_Data, TAG_END);
					break;

				case MUIA_Debug_LogMode:
					SetAttrs(obj, MUIA_Debug_LogMode, tag->ti_Data, TAG_END);
					break;

				case MUIA_Debug_DumpMode:
					SetAttrs(obj, MUIA_Debug_DumpMode, tag->ti_Data, TAG_END);
					break;

				case MUIA_Debug_LogFile:
					SetAttrs(obj, MUIA_Debug_LogFile, tag->ti_Data, TAG_END);
					break;

				case MUIA_Debug_DumpFile:
					SetAttrs(obj, MUIA_Debug_DumpFile, tag->ti_Data, TAG_END);
					break;
			}
		}

		/* notifications */

		//DoMethod(data->m_sl_current_hit, MUIM_Notify, MUIA_Pressed, FALSE, obj, 1, MUIM_Debug_ShowHit);
		 DoMethod(data->m_sl_current_hit, MUIM_Notify, MUIA_Slider_Level, MUIV_EveryTime, obj, 1, MUIM_Debug_ShowHit);

		DoMethod(data->m_bt_analyze, MUIM_Notify, MUIA_Pressed, FALSE, obj, 1, MUIM_Debug_Analyze);


		SetAttrs(data->m_st_ramdebug_file, MUIA_Disabled, TRUE, TAG_END);
		SetAttrs(data->m_st_dump_file, MUIA_Disabled, TRUE, TAG_END);

		SetAttrs(data->m_po_ramdebug_file, MUIA_Disabled, TRUE, TAG_END);
		SetAttrs(data->m_po_dump_file, MUIA_Disabled, TRUE, TAG_END);


		DoMethod(data->m_cy_ramdebug_mode, MUIM_Notify, MUIA_Cycle_Active, LOG_RAMDEBUG, data->m_st_ramdebug_file, 3, MUIM_Set, MUIA_Disabled, TRUE);
		DoMethod(data->m_cy_ramdebug_mode, MUIM_Notify, MUIA_Cycle_Active, LOG_RAMDEBUG, data->m_po_ramdebug_file, 3, MUIM_Set, MUIA_Disabled, TRUE);
		DoMethod(data->m_cy_ramdebug_mode, MUIM_Notify, MUIA_Cycle_Active, LOG_FILE, data->m_st_ramdebug_file, 3, MUIM_Set, MUIA_Disabled, FALSE);
		DoMethod(data->m_cy_ramdebug_mode, MUIM_Notify, MUIA_Cycle_Active, LOG_FILE, data->m_po_ramdebug_file, 3, MUIM_Set, MUIA_Disabled, FALSE);


		DoMethod(data->m_cy_dump_mode, MUIM_Notify, MUIA_Cycle_Active, DUMP_AUTO, data->m_st_dump_file, 3, MUIM_Set, MUIA_Disabled, TRUE);
		DoMethod(data->m_cy_dump_mode, MUIM_Notify, MUIA_Cycle_Active, DUMP_AUTO, data->m_po_dump_file, 3, MUIM_Set, MUIA_Disabled, TRUE);
		DoMethod(data->m_cy_dump_mode, MUIM_Notify, MUIA_Cycle_Active, DUMP_FILE, data->m_st_dump_file, 3, MUIM_Set, MUIA_Disabled, FALSE);
		DoMethod(data->m_cy_dump_mode, MUIM_Notify, MUIA_Cycle_Active, DUMP_FILE, data->m_po_dump_file, 3, MUIM_Set, MUIA_Disabled, FALSE);


		DoMethod(data->m_lv_hits, MUIM_Notify, MUIA_Listview_DoubleClick, TRUE, obj, 1, MUIM_Debug_PerformAction);
	}

	return (ULONG)obj;
}

static ULONG DebugDispose(struct IClass *cl, Object *obj, struct opSet *msg)
{
	struct DebugData *data;

	data = (struct DebugData *)INST_DATA(cl, obj);

	return (DoSuperMethodA(cl, obj, msg));
}


static ULONG DebugSet(struct IClass *cl, Object *obj, struct opSet *msg)
{
	struct DebugData *data;
	struct TagItem *tags, *tag;

	data = (struct DebugData *)INST_DATA(cl, obj);

	for (tags=((struct opSet *)msg)->ops_AttrList;tag=NextTagItem(&tags);)
	{
		switch (tag->ti_Tag)
		{
			case MUIA_Debug_LogMode:
				data->m_ramdebug_mode = tag->ti_Data;
				SetAttrs(data->m_cy_ramdebug_mode, MUIA_Cycle_Active, data->m_ramdebug_mode, TAG_END);
				break;

			case MUIA_Debug_DumpMode:
				data->m_ramdebug_mode = tag->ti_Data;
				SetAttrs(data->m_cy_dump_mode, MUIA_Cycle_Active, data->m_dump_mode, TAG_END);
				break;

			case MUIA_Debug_Executable:
				if(tag->ti_Data)
				{
					strcpy(data->m_executable, (char *) tag->ti_Data);
					SetAttrs(data->m_st_executable, MUIA_Text_Contents, data->m_executable, TAG_END);
				}
				break;

			case MUIA_Debug_LogFile:
				if(tag->ti_Data)
				{
					strcpy(data->m_ramdebugfile, (char *) tag->ti_Data);
					SetAttrs(data->m_st_ramdebug_file, MUIA_Text_Contents, data->m_ramdebugfile, TAG_END);
				}
				break;

			case MUIA_Debug_DumpFile:
				if(tag->ti_Data)
				{
					strcpy(data->m_dumpfile, (char *) tag->ti_Data);
					SetAttrs(data->m_st_dump_file, MUIA_Text_Contents, data->m_dumpfile, TAG_END);
				}
				break;

			case MUIA_Debug_CurrentHit:
				data->m_current_hit = tag->ti_Data;
				SetAttrs(data->m_sl_current_hit, MUIA_Slider_Level, data->m_current_hit, TAG_END);
				break;
		}
	}

	return (DoSuperMethodA(cl, obj, msg));
}

static ULONG Debug_ShowHit(struct DebugData *data);

static ULONG Debug_Analyze(struct DebugData *data)
{
	char *ptr;
	int val;

	string logfilename = "t:log";
	string dumpfilename= "t:dump";
	char buffer[1024];

	GetAttr(MUIA_Text_Contents, data->m_st_executable, (ULONG *) &ptr);
	strcpy(data->m_executable, ptr);

	GetAttr(MUIA_Cycle_Active, data->m_cy_ramdebug_mode, (ULONG *) &val);
	data->m_ramdebug_mode = val;

	GetAttr(MUIA_Cycle_Active, data->m_cy_dump_mode, (ULONG *) &val);
	data->m_dump_mode = val;

	GetAttr(MUIA_Text_Contents, data->m_st_ramdebug_file, (ULONG *) &ptr);
	strcpy(data->m_ramdebugfile, ptr);

	GetAttr(MUIA_Text_Contents, data->m_st_dump_file, (ULONG *) &ptr);
	strcpy(data->m_dumpfile, ptr);

	if(data->m_ramdebug_mode == LOG_RAMDEBUG)
	{
		snprintf(buffer, 1024, "getramdebuglog >nil: %s", logfilename.c_str());
		Execute(buffer, 0, 0);
	}
	else
	{
		logfilename = data->m_ramdebugfile;
	}

	if(data->m_dump_mode == DUMP_AUTO )
	{
		snprintf(buffer, 1024, "objdump -d -r %s >%s", data->m_executable, dumpfilename.c_str());
		strcpy(data->m_dumpfile, dumpfilename.c_str());
		Execute(buffer, 0, 0);
	}
	else
	{
		dumpfilename =  data->m_dumpfile;
	}

	data->m_hits.clear();
	data->m_funcs.clear();

	data->m_hits = GetHits(logfilename);
	data->m_funcs = GetFuncs(dumpfilename);

	if(data->m_hits.size())
	{
		SetAttrs(data->m_sl_current_hit, MUIA_Disabled, FALSE, TAG_END);
		SetAttrs(data->m_sl_current_hit, MUIA_Slider_Max, data->m_hits.size(), TAG_END);

		Debug_ShowHit(data);
	}
	else
	{
		SetAttrs(data->m_sl_current_hit, MUIA_Disabled, TRUE, TAG_END);
	}

	return TRUE;
}

static ULONG Debug_ShowHit(struct DebugData *data)
{
	char * executable = FilePart(data->m_executable);
	unsigned int index;
	long line;
	int val;
	struct Hit hit;

	GetAttr(MUIA_Slider_Level, data->m_sl_current_hit , (ULONG *) &val);
	data->m_current_hit = val-1;

	DoMethod(data->m_lv_hits, MUIM_List_Clear);

	if(data->m_current_hit < data->m_hits.size())
	{
		hit = data->m_hits[data->m_current_hit];

		if(strstr(hit.m_srr0.m_name.c_str(), executable))
		{
			if(GetInfoForOffset(data->m_dumpfile, hit.m_srr0.m_offset, data->m_funcs, &line, &index))
			{
				struct LI_Data item;

				snprintf(item.m_register, 128, "SRR0");
				snprintf(item.m_function_name, 128, "<%s>", data->m_funcs[index].m_name.c_str());
				snprintf(item.m_offset, 128, "0x%lx", hit.m_srr0.m_offset);
				snprintf(item.m_line, 128, "%ld", line);

				DoMethod(data->m_lv_hits, MUIM_List_InsertSingle, &item, MUIV_List_Insert_Bottom);
			}

		}

		if(strstr(hit.m_lr.m_name.c_str(), executable))
		{
			if(GetInfoForOffset(data->m_dumpfile, hit.m_lr.m_offset, data->m_funcs, &line, &index))
			{
				struct LI_Data item;

				snprintf(item.m_register, 128, "LR");
				snprintf(item.m_function_name, 128, "<%s>", data->m_funcs[index].m_name.c_str());
				snprintf(item.m_offset, 128, "0x%lx", hit.m_lr.m_offset);
				snprintf(item.m_line, 128, "%ld", line);

				DoMethod(data->m_lv_hits, MUIM_List_InsertSingle, &item, MUIV_List_Insert_Bottom);
			}
		}

		for(unsigned int j=0; j<hit.m_stackframe.size(); j++)
		{
			if(strstr(hit.m_stackframe[j].m_name.c_str(), executable))
			{
				if(GetInfoForOffset(data->m_dumpfile, hit.m_stackframe[j].m_offset, data->m_funcs, &line, &index))
				{
					struct LI_Data item;

					snprintf(item.m_register, 128, "LR[%ld]", hit.m_stackframe[j].m_depth);
					snprintf(item.m_function_name, 128, "<%s>", data->m_funcs[index].m_name.c_str());
					snprintf(item.m_offset, 128, "0x%lx", hit.m_stackframe[j].m_offset);
					snprintf(item.m_line, 128, "%ld", line);

					DoMethod(data->m_lv_hits, MUIM_List_InsertSingle, &item, MUIV_List_Insert_Bottom);
				}
			}
		}
	}

	return TRUE;
}

static ULONG Debug_PerformAction(struct DebugData *data)
{
	int val;
	struct LI_Data * item;
	char buffer[1024];

	GetAttr(MUIA_List_Active, data->m_lst_hits, (ULONG *) &val);
	
	if(item)
	{
		DoMethod(data->m_lv_hits, MUIM_List_GetEntry, val, &item);
		snprintf(buffer, 1024, "rx gotoline.rexx %s %ld", data->m_dumpfile, strtol(item->m_line, NULL, 10));   
		Execute(buffer, 0, 0);
	}

	return TRUE;
}

DISPATCHERPROTO(DebugDispatcher)
{
	DISPATCHERARG

	struct DebugData *data = (struct DebugData *)INST_DATA(cl, obj);

	switch(msg->MethodID){
		case OM_NEW :						return DebugNew(cl, obj, (struct opSet *) msg);
		case OM_DISPOSE :					return DebugDispose(cl, obj, (struct opSet *)msg);
		case OM_SET:						return DebugSet(cl, obj, (struct opSet *)msg);

		case MUIM_Debug_ShowHit :		return Debug_ShowHit(data);
		case MUIM_Debug_PerformAction :	return Debug_PerformAction(data);
		case MUIM_Debug_Analyze :		return Debug_Analyze(data);
	}
		return DoSuperMethodA(cl, obj, msg);
}

Object * OpenMainWindow(void)
{
	Object *obj_debug;

	app = (Object *)ApplicationObject,
		MUIA_Application_Author, "Fabien",
		MUIA_Application_Base, "DEBUG",
		MUIA_Application_Title, "Debug GUI",
		MUIA_Application_Version, "$VER: Debug 0.1 (14/03/05)",
		MUIA_Application_Copyright, "Fabien Coeurjoly",
		MUIA_Application_Description, "Debug Tool",
		MUIA_Application_HelpFile, NULL,

		SubWindow, window = WindowObject,
			MUIA_Window_Title, "Debug",
			MUIA_Window_ID, MAKE_ID('D', 'E', 'B', '1'),
			WindowContents,
				VGroup,
						Child, obj_debug = (Object *) NewObject(cl_debugo->mcc_Class, NULL, MUIA_Debug_Executable, "hitmania.db", TAG_DONE),
				End,
		End,
	End;

	if (app)
	{
	
		DoMethod(window,
			MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
			app, 2,
			MUIM_Application_ReturnID, MUIV_Application_ReturnID_Quit);

		SetAttrs(window, MUIA_Window_Open, TRUE, TAG_END);
	}

	return app;
}

int Initialize(void)
{
	int res = 1;

	IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 39L);
	MUIMasterBase = OpenLibrary(MUIMASTER_NAME, MUIMASTER_VMIN);
	UtilityBase = OpenLibrary("utility.library", 39);

	if (IntuitionBase == NULL)
	{
		printf("Can't open 'intuition.library' V39\n");
		res = 0;
	}
	if (MUIMasterBase == NULL)
	{
		printf("Can't open '%s' V%d\n", MUIMASTER_NAME, MUIMASTER_VMIN);
		res = 0;
	}
	if (UtilityBase == NULL)
	{
		printf("Can't open 'utility.library' V39\n");
		res = 0;
	}

	cl_debugo = MUI_CreateCustomClass(NULL, MUIC_Group, NULL, sizeof(struct DebugData), &DebugDispatcher);
	if (cl_debugo == NULL)
	{
		printf("Can't create debug internal class\n");
		res = 0;
	}

	return res;
}

void DeInitialize(void)
{
	MUI_DeleteCustomClass(cl_debugo);
	CloseLibrary(UtilityBase);
	CloseLibrary(MUIMasterBase);
	CloseLibrary((struct Library *)IntuitionBase);
}


int main(int argc, char **argv)
{
	int res = 0;

	if (Initialize())
	{
		app = OpenMainWindow();
		if (app)
		{
			ULONG sigs = 0;

			while (DoMethod(app,MUIM_Application_NewInput,&sigs) != ((unsigned int) MUIV_Application_ReturnID_Quit))
			{
				if (sigs)
				{
					sigs = Wait(sigs | SIGBREAKF_CTRL_C);
					if (sigs & SIGBREAKF_CTRL_C) break;
				}
			}

			SetAttrs(window, MUIA_Window_Open, FALSE, TAG_END);
			MUI_DisposeObject(app);
		}
		else
		{
			res = 2;
		}
	}
	else
	{
		res = 1;
	}

	DeInitialize();

	return res;
}
