/*************************************************************************/
/*                                                                       */
/* Requester Classe (Window SubClass) for handle async requester         */
/*                                                                       */
/*************************************************************************/

#ifndef NDEBUG
#define NDEBUG
#endif

#include "global.h"
#include "array.h"
#include "requester.h"
#include "gui.h"
#include "guigfx.h"
#include "locale.h"

extern struct Library *MUIMasterBase;
extern struct MUI_CustomClass *mcc_requester, *mcc_guigfx;
extern Object *MainApp;

/***********************************************************************/

struct Req
{
    array_t  ArrayButton; // Array of Button struct
    BOOL     AppSleep;    // Sleep app ?
    Object  *WinSleep;    // Object for window sleep only
};

struct ReqButton
{
    char   *Name;       // Name of the button
    Object *Obj;
    array_t ArrayTags;  // Array of ULONG
};

/***********************************************************************/

static ULONG mNew(struct IClass *cl,Object *obj,Msg msg)
{
	struct Req       tmp = { NULL, TRUE, NULL };
    struct ReqButton *button=NULL,*lastbut=NULL;
    struct TagItem   *tags,*tag;
	Object           *grp,*but,*img=NULL;
	char             *Title=NULL,*Body=NULL,*Icon=NULL;
	ULONG            *rtag=NULL,*stag=NULL,i=1,n;

    /* parse initial taglist */
    PUTS("Requester New: Read tags...");

    for (tags=((struct opSet *)msg)->ops_AttrList;(tag=NextTagItem(&tags));)
    {
        PF("Read tag: %lx\n",tag->ti_Tag);
        switch (tag->ti_Tag)
        {
            case MUIA_Req_AppSleep:
                PUTS("AppSleep");
                tmp.AppSleep=tag->ti_Data;
                break;
            case MUIA_Req_WinSleep:
                PUTS("WinSleep");
                tmp.WinSleep=(Object *)tag->ti_Data;
                break;
            case MUIA_Req_Title:
                PUTS("Title");
                if (!tag->ti_Data) return (ULONG)NULL;
                Title=(char *)tag->ti_Data;
                break;
            case MUIA_Req_Body:
                PUTS("Body");
                if (!tag->ti_Data) return (ULONG)NULL;
                Body=(char *)tag->ti_Data;
                break;
			case MUIA_Req_Icon:
				if (tag->ti_Data) MyStrCpy(&Icon,(char *)tag->ti_Data);
				break;
            case MUIA_Req_Button:
                if (!tmp.ArrayButton)
                {
                    // First create array
                    PUTS("First button");
                    tmp.ArrayButton=Array_New(sizeof(struct ReqButton),2);
                    if (!tmp.ArrayButton) return (ULONG)NULL;
                }
                // Add button
                if (!button)
                {
                    PUTS("Add Button");
                    button=Array_Add(tmp.ArrayButton);
                    if ( (button) && (tag->ti_Data) )
                    {
                        MyStrCpyOnly(&(button->Name),(char *)tag->ti_Data);
                        button->ArrayTags=Array_New(sizeof(ULONG),2);
                        if (!button->ArrayTags) PUTS("## No button arraytags!");
						
                    }
                    else PF("## no array add (%08lx,%08lx)\n",button,tag->ti_Data);
                }
                else PUTS("## Req button already created, check all!");
                break;
            case MUIA_Req_Object:
				button->Obj=(Object *)tag->ti_Data;
				if ((tmp.ArrayButton) && (button) && (button->ArrayTags) && (tag->ti_Data) )
                {
                    // Add method to current button
					PF("Add Object: %08lx (App:%08lx)\n",button->Obj,MainApp);
					rtag=Array_Add(button->ArrayTags);
					*rtag=MUIM_Application_PushMethod;
					rtag=Array_Add(button->ArrayTags);
					if (rtag) *rtag=(ULONG)button->Obj;
					stag=Array_Add(button->ArrayTags);
					if (stag) *stag=0;
					i=0;
                }
				else PF("## Req Object NULL here (%08lx,%08lx,%08lx)\n",tmp.ArrayButton,button,button->ArrayTags);
                break;
            case MUIA_Req_Tag:
                if ((tmp.ArrayButton) && (button) && (button->ArrayTags) && (button->Obj) )
                {
                    // Add tag to current button
					PF("Add Tag: %08lx\n",tag->ti_Data);
                    rtag=Array_Add(button->ArrayTags);
					if (rtag)
					{
					    *rtag=tag->ti_Data;
						i=i+1;
					}
                }
                else PF("## Invalid Req Tag here (%08lx,%08lx,%08lx,%08lx)",tmp.ArrayButton,button,button->ArrayTags,button->Obj);
                break;
            case MUIA_Req_Next:
				if ( (tmp.ArrayButton) && (button) && (button->Obj) && (button->ArrayTags) )
                {
                    // End of button
					rtag=Array_Add(button->ArrayTags); 
					*rtag=0L;
					for (n=2,rtag=Array_First(button->ArrayTags);(n) && (rtag);n--)
					{
						rtag=Array_Next(button->ArrayTags,rtag);
					}
					PF("rtag: %08lx\n",*rtag);
					if (rtag) *rtag=i;
                    PUTS("End of button");
					PF("Created: *stag=%ld size=%ld\n",(stag) ? *stag : -1, Array_GetSize(button->ArrayTags));
#ifndef NDEBUG
					PF("DoMethod(%08lx",MainApp);
					for (rtag=Array_First(button->ArrayTags);(rtag);rtag=Array_Next(button->ArrayTags,rtag))
					{
						PF(",%08lx",*rtag);
					}
					PUTS(");\n");
#endif
					button=NULL;
					rtag=NULL;
					stag=NULL;
				}
                else PF("## Invalid tag Next (%08lx,%08lx)",tmp.ArrayButton,button);
                break;
        }
    }

    PUTS("Build Window");

    ASSERT(Title);
    ASSERT(Body);
    ASSERT(tmp.ArrayButton);
	if (!Icon) MyStrCpy(&Icon,"PROGDIR:images/req.png");
	if (Icon) img=NewObject(mcc_guigfx->mcc_Class, NULL,
				  MUIA_GG_File,   Icon,
				  TAG_DONE);
	free(Icon);
	Icon=NULL;

    obj = (Object *)DoSuperNew(cl,obj,
        MUIA_Window_Title,       Title,
        MUIA_Window_CloseGadget, TRUE,
        MUIA_Window_SizeGadget,  FALSE,
        WindowContents, VGroup,
            MUIA_Background, MUII_RequesterBack,
            Child, HGroup,
                MUIA_Frame,      MUIV_Frame_Group,
                MUIA_Background, MUII_GroupBack,
				Child, (img) ? img : Label( LOC(LOC_REQ_ICON_NOT_LOADED,"Icon Not\nLoaded!")),
                Child, TextObject,
                    MUIA_Font,          MUIV_Font_Big,
                    MUIA_Text_Contents, Body,
                    End,
                End,
            Child, MUI_NewObject(MUIC_Rectangle,
                MUIA_Rectangle_HBar, TRUE,
                MUIA_FixHeight,      10,
                End,
            Child, grp=HGroup,
                Child, HSpace(0),
                End,
            End,
        End;

    if (obj)
    {
        
        /* Cration russie */
		struct Req *data = INST_DATA(cl,obj);
        PUTS("Object not NULL");
        *data = tmp;

        // Fill buttons
        for( button=Array_First(data->ArrayButton);(button);button=Array_Next(data->ArrayButton,button) )
        {
            // Add button
            PF("Add ReqButton '%s'\n",button->Name);
            but=SimpleButton(button->Name);
            DoMethod(grp, OM_ADDMEMBER, but);
            DoMethod(grp, OM_ADDMEMBER, HSpace(0));
            DoMethod(but, MUIM_Notify, MUIA_Pressed, 0,
                     obj, 2,
                     MUIM_Req_Close, button);
            set(but, MUIA_CycleChain, TRUE);
            if (!lastbut) set(obj, MUIA_Window_ActiveObject, but);
            lastbut=button;
        }

        DoMethod(obj, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
                 obj, 2,
                 MUIM_Req_Close, lastbut);

        DoMethod(MainApp,OM_ADDMEMBER,obj);
        if (data->WinSleep)
        {
            set(obj, MUIA_Window_RefWindow, data->WinSleep);
            set(data->WinSleep,MUIA_Window_Sleep, TRUE);
        }
        if (data->AppSleep)
        {
            PUTS("Sleep GUI");
            set(MainApp, MUIA_Application_Sleep, TRUE);
            set(obj,MUIA_Window_Sleep, FALSE);
        }
		if (xget(MainApp, MUIA_Application_Iconified)==TRUE) set(MainApp, MUIA_Application_Iconified, FALSE);
        set(obj, MUIA_Window_Open, TRUE);
        if (xget(obj, MUIA_Window_Open)==0)
        {
            PUTS("## Can't open Requester window!!");
        }
        
    }
    else fail("Objet not created in class Requester!\n");
    return (ULONG)obj;
}

static ULONG mDispose(struct IClass *cl, Object *obj, Msg msg)
{
    struct Req       *data = INST_DATA(cl, obj);
    struct ReqButton *button;
    if (data->AppSleep)
    {
        PUTS("UnSleep GUI");
        set(MainApp, MUIA_Application_Sleep, FALSE);
    }
    if (data->WinSleep)
    {
        set(data->WinSleep,MUIA_Window_Sleep, FALSE);
    }
    PUTS("Free Requester datas...");
    if (data->ArrayButton)
    {
        for( button=Array_First(data->ArrayButton);(button);button=Array_Next(data->ArrayButton, button) )
        {
			if (button->Name)      free(button->Name);
            if (button->ArrayTags) Array_Free(button->ArrayTags);
        }
        Array_Free(data->ArrayButton);
    }
    PUTS("Ok");
    return DoSuperMethodA(cl,obj,msg);
}

/***********************************************************************/

static ULONG mClose(struct IClass *cl,Object *obj, struct MUIP_Req_Close *msg)
{
    struct Req *data = INST_DATA(cl,obj);
    struct ReqButton *button;
	ULONG *tags;

    PF("Requester Close Method: %08lx\n",msg->button);

    set(obj, MUIA_Window_Open, FALSE);
	DoMethod(MainApp, MUIM_Application_InputBuffered);
    DoMethod(MainApp, MUIM_Application_PushMethod,
             obj, 1,
             OM_DISPOSE);
	if (msg->button)
    {
        button=msg->button;
        PF("Pressed: '%s'\n",button->Name);
        // DoMethod for the button
        if (button->Obj)
        {
			#ifndef NDEBUG
			PF("DoMethod(%08lx",MainApp);
			for (tags=Array_First(button->ArrayTags);(tags);tags=Array_Next(button->ArrayTags,tags))
			{
				PF(",%08lx",*tags);
			}
			PUTS(");\n");
			#endif
			tags=Array_First(button->ArrayTags);
			DoMethodA(MainApp, (Msg)tags);
					  //  MUIM_Application_PushMethod,
					  //  button->Obj, 2;
        }
        else PUTS("No Object-> No Method :)");
    }
	DoMethod(MainApp, OM_REMMEMBER,obj);
    return TRUE;
}

/***********************************************************************/

#ifdef __MORPHOS__
static ULONG RealDispatcher( void );
static struct EmulLibEntry Dispatcher = { TRAP_LIB, 0, (void(*)(void))RealDispatcher };
static ULONG RealDispatcher( void )
{
    struct IClass *cl = (struct IClass *)REG_A0;
    Object *obj = (Object *)REG_A2;
    Msg msg = (Msg)REG_A1;
#else
static ASM ULONG Dispatcher(  REG(a0,struct IClass *cl), REG(a2,Object *obj), REG(a1,Msg msg) )
{
#endif
    //PF("BM: %lX\n",msg->MethodID);
    switch (msg->MethodID)
    {
        case OM_NEW:                 return           mNew(cl,obj,(APTR)msg);
        case OM_DISPOSE:             return       mDispose(cl,obj,(APTR)msg);
        case MUIM_Req_Close:         return         mClose(cl,obj,(APTR)msg);
        default:                     return DoSuperMethodA(cl,obj,(APTR)msg);
    }
}

/***************************************************************************/

BOOL ASM initRequester(void)
{
    if ((mcc_requester = MUI_CreateCustomClass(NULL,MUIC_Window,NULL,sizeof(struct Requester),(APTR)&Dispatcher)))
    {
        PUTS("MCC Requester created");
        if (MUIMasterBase->lib_Version>=20)
            mcc_requester->mcc_Class->cl_ID = "PolyNetRequester";
        return TRUE;
    }
    PUTS("MCC Requester not created!");
    return FALSE;
}

/***********************************************************************/
