struct Data {

	/* background randomizer input handler
	 */
	ULONG bgrandom_timer_count;
	ULONG bgrandom_timer_wait;

	ULONG bgrandom_inputhandler_added;
	struct MUI_InputHandlerNode bgrandom_ihnode;

	/* cache cleanup counter
	 */
	ULONG cachecleanup_timer_count;
	ULONG cachecleanup_timer_wait;

	ULONG cachecleanup_inputhandler_added;
	struct MUI_InputHandlerNode cachecleanup_ihnode;

};


DEFNEW
{
	struct Data *data;

	obj = DoSuperNew(cl, obj,
		MUIA_Application_Title,       APPNAME,
		MUIA_Application_Version,     "$VER: "APPNAME" "LVERTAG,
		MUIA_Application_Copyright,   " 2001-2005 by David Gerber,  2005-" COPYRIGHTYEAR_END " by Ambient Open Source Team, All Rights Reserved.",
		MUIA_Application_Author,      "David Gerber, Ambient Open Source Team",
		MUIA_Application_UsedClasses, classlist,
		MUIA_Application_Description, "The MorphOS desktop",
		MUIA_Application_Base,        "AMBIENT",
		MUIA_Application_SingleTask,  TRUE,
		MUIA_Application_UseRexx,     FALSE,
		MUIA_Application_NoIconify,   TRUE,
		TAG_MORE, INITTAGS
	);

	if (!obj)
	{
		return (0);
	}

	data = INST_DATA(cl, obj);


   ...


	/* cache cleanup handler (invoke cleanup routine every 30 minutes) */

	data->cachecleanup_timer_count = 30;
	data->cachecleanup_timer_wait = 30;

	data->cachecleanup_ihnode.ihn_Object = obj;
	data->cachecleanup_ihnode.ihn_Flags = MUIIHNF_TIMER;
	data->cachecleanup_ihnode.ihn_Millis = 60 * 1000;
	data->cachecleanup_ihnode.ihn_Method = MM_Application_CleanupCache;
	DoMethod(obj, MUIM_Application_AddInputHandler, &data->cachecleanup_ihnode);
	data->cachecleanup_inputhandler_added = TRUE;

	return ((ULONG)obj);
}


DEFDISP
{
	GETDATA;


	if(data->bgrandom_inputhandler_added)
	{
		DoMethod(obj, MUIM_Application_RemInputHandler, &data->bgrandom_ihnode);
	}

	if(data->cachecleanup_inputhandler_added)
	{
		DoMethod(obj, MUIM_Application_RemInputHandler, &data->cachecleanup_ihnode);
	}


	return (DOSUPER);
}


DEFGET
{
}

DEFSET
{
	GETDATA;

	FORTAG(INITTAGS)
	{
        ...

		case MA_Application_BackgroundRefreshDelay:
		{
			if(data->bgrandom_inputhandler_added)
			{
				DoMethod(obj, MUIM_Application_RemInputHandler, &data->bgrandom_ihnode);
				data->bgrandom_inputhandler_added = FALSE;
			}

			if(tag->ti_Data)
			{
				data->bgrandom_timer_wait = data->bgrandom_timer_count = (ULONG)tag->ti_Data;

				data->bgrandom_ihnode.ihn_Object = obj;
				data->bgrandom_ihnode.ihn_Flags = MUIIHNF_TIMER;
				data->bgrandom_ihnode.ihn_Millis = 60*1000;
				data->bgrandom_ihnode.ihn_Method = MM_Application_ReloadBackgrounds;

				DoMethod(obj, MUIM_Application_AddInputHandler, &data->bgrandom_ihnode);

				data->bgrandom_inputhandler_added = TRUE;

			}
		}
	}
	NEXTTAG

	return (DOSUPER);
}

DEFTMETHOD(Application_CleanupCache)
{
	GETDATA;

	if(--data->cachecleanup_timer_count == 0)
	{
		data->cachecleanup_timer_count = data->cachecleanup_timer_wait;

		/*
		 * TODO: Make this one asynchronous? Probably good idea...
		 */

		cache_deleteunused();

	}

	return (0);
}
